from os.path import join, isfile, exists, dirname
from time import sleep

### Data analysis:
from inspect import CO_ASYNC_GENERATOR
from sre_constants import CATEGORY_DIGIT
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import numpy as np
from math import *

from epics import caget, caput

import  iexcode.instruments.cfg as iex

from iexcode.instruments.xrays import *
from iexcode.instruments.xrays import _energy_range_min_max
from iexcode.instruments.current_amplifiers import current2flux
from iexcode.instruments.diagnostics import _diagnostics_dict, diagnostics_all_out, diagnostics_all_in,diode_c,diode_d
from iexcode.instruments.m3r import *
from iexcode.instruments.slits import *
from iexcode.instruments.utilities import *
from iexcode.instruments.Logfile import log_name_set,log_print
from iexcode.macros.quick_plot import *

##############################################################################################################
################################          ID calibration scripts        ##############################
##############################################################################################################

######  Energy range follows ID ########

def ID_calibration_scan(ID_start,ID_stop,ID_step,bandwidth=10,QP=None,Harm=1,scanIOC=None):
    """
    Sets the ID_SP and scans the mono:
        start,stop,step are for ID
    if QP is not None adjusts the mono range to offset accordingly
    !!! DOES NOT CHANGE QP VALUE !!!!
    !!! Doesn't put diodes in !!!
    """
    
    print("""
             !!! DOES NOT CHANGE QP VALUE !!!!
             !!! Doesn't put diodes in !!!""")
    
    log_name_set(iex.BL.endstation,filename_suffix='IDCalibration')

    #Getting mono max range based on grating
    grt_density = mono_grating_density_get()
    c=grt_density/1200            # c = 1 for MEG, 2 for HEG
    if c == 1: 
        maxhv = 3000
    elif c == 2: 
        maxhv = 2000
   
    log_print(comment="====== Starting Mono Scan vs ID ======")

    for ID_energy in range_up(ID_start,ID_stop,ID_step):
        print("\n------------------  ID-SP @ "+str(ID_energy)+"   ------------------")
        mono_energy_set(min(ID_energy,2999))
        ID_SP_set(ID_energy)
        apertures_set()
        branch = m3r_branch()
        if branch == 'd':
            m3r_align()
        if QP is None:
            start_hv=min(ID_energy-ID_energy/(bandwidth*1.0)*Harm,maxhv)  # FR update range 11/20/2019
            stop_hv=min(ID_energy+ID_energy/(bandwidth*2.0)*Harm,maxhv)
            step_hv=round(ID_energy/300.0,1)
#            start_hv=min(ID_energy-ID_energy/(bandwidth*2.0)*Harm,maxhv)  # shift the window by doing -BW*2/+BW*1 instead of -BW*1'/'+BW*2
#            stop_hv=min(ID_energy+ID_energy/(bandwidth*1.0)*Harm,maxhv)
#            step_hv=round(ID_energy/500.0,1)
        else:
            start_hv=min(ID_energy-ID_energy/(bandwidth*2.5)*Harm,maxhv)
            stop_hv=min(ID_energy+ID_energy/(bandwidth*0.7)*Harm,maxhv)
            step_hv=round(ID_energy/300.0,1)

        mono_scan_fillin(start_hv,stop_hv,step_hv)
        iex.BL.mda.go()
        sleep(1)

        FileNum  = iex.BL.mda.lastFileNum()
        if ID_energy == ID_start:
            FirstFileNum = FileNum     # Record First Scan Number
    
    LastFileNum = FileNum              # Record Last Scan Number

    print("Use:  update_id_dict("+str(FirstFileNum)+","+str(LastFileNum)+"ignore=[] ,update_file,path=None,prefix='ARPES',q=True)")



def ID_calibration_mode(ID_mode,start=None,stop=None,step=25,Harm=1,bandwidth=10):
    '''
    Runs full calibration with a 25 eV step (by default) for a given mode:
            which = 'RCP','LCP', 'V','H'
    Better to run each mode in a separated cell (print statement at the end helps with calibration) 
    '''
    branch = m3r_branch()
    diagnostics_all_out(diode_stay_in=True)

    polarization(ID_mode)

    grt = mono_grating_get() 

    QP_ratio = ID_QP_ratio_get()
    
    hv_min,hv_max = _energy_range_min_max()
    if start == None:
        start = hv_min
    if stop == None:
        stop=hv_max-50

    if QP_ratio <100:
        stop=min(stop,1500)
        QP_ratio=1
    else: 
        QP_ratio=None

    ID_calibration_scan(start,stop,step,bandwidth=bandwidth,QP=QP_ratio,Harm=Harm)

###############################################################################################
######################################### IDCalibration New ###################################
###############################################################################################



def id_calibration_fit(first_scan,last_scan,det,PolyRank,**kwargs):
    
    
    #id_calibration_fit(FirstLast[0],FirstLast[1],det,poly_order,mytitle=mytitle,countby=countby,plotMin=plt_min,plotMax=plt_max,plotType=plotType,filepath=filepath,prefix=prefix)
    """
    Fits calibration curves fpr each GRTs and modes included in the data set. Creates 3 dictionnaries:
    
        coefs={[GRT][mode][[breakpoint1,[coefs]],[breakpoint2,[coefs]...}
        xdata= {[GRT][mode][[breakpoint1,[flux curve x axis]],[breakpoint2,[flux curve x axis]...}
        fdata= {[GRT][mode][[breakpoint1,[flux curve y axis]],[breakpoint2,[flux curve y axis]...}
        
    kwargs:
        countby = 1 by default
        mytitle = '' by default
        plotMin =  min x range for plotting (default 250)
        plotMax =  max x range for plotting (default 3000)
        plotType = ['dif,'fit,'flux'], full set by default
        filepath = None,  uses current mda filepath unless specified
              e.g. user : filepath='/net/s29data/export/data_29idc/2018_2/UserName/mda/'
              e.g. staff: filepath='/net/s29data/export/data_29idb/2018_2/mda/'
        prefix = None, uses current mda prefix unless specified 
        scanIOC = None, uses BL_ioc() unless specified
    """

    kwargs.setdefault('countby',1)
    kwargs.setdefault('mytitle','')
    #kwargs.setdefault('plotMin',2000)
    kwargs.setdefault('plotMin',225)
    kwargs.setdefault('plotMax',3000)
    kwargs.setdefault('plotType',['fit','dif','flux'])
    kwargs.setdefault('filepath',iex.BL.mda.filepath())
    kwargs.setdefault('prefix',iex.BL.mda.prefix())

    countby=kwargs['countby']
    mytitle=kwargs['mytitle']
    plotMin=kwargs['plotMin']
    plotMax=kwargs['plotMax']
    filepath=kwargs['filepath']
    prefix=kwargs['prefix']
    

    def calibration_curve(first,last,det,countby,filepath,prefix):
        mono_list=[]
        ID_list=[]
        max_list=[]
        flux_list=[]
        print("filepath = ",filepath)
        print("prefix = ",prefix)
        print("First scan = ",first)
        print("Last scan  = ",last)
        for i in range(first,last,countby):
            x,y,x_name,y_name=mda_1D(i,det,1,0,filepath,prefix)   #mda_1D(ScanNum,DetectorNum,coeff=1,bckg=0,filepath=None,prefix=None,scanIOC=None):
            v,w,v_name,w_name=mda_1D(i, 4 ,1,0,filepath,prefix)
            if y != []:
                n=y.index(max(y))        # finds the peak max intensity index
                e=round(x[n],2)            # finds the corresponding mono energy
                sp=round(w[2]*1000,0)   # reads the ID set point
                mono_list.append(e)
                max_list.append(max(y))
                ID_list.append(sp)
                flux_list.append(current2flux(max(y),e,p=None))
        return mono_list,ID_list,max_list,flux_list
    
    Mono_max,ID_SP,int_max,flux=calibration_curve(first_scan,last_scan+1,det,countby,filepath,prefix)
    nPt_gaus=200
    x_HR =  np.linspace(Mono_max[0], Mono_max[-1], nPt_gaus)
    
    # Data
    xdata = np.array(Mono_max)
    ydata = np.array(ID_SP)
    zdata = np.array(int_max)
    fdata = np.array(flux)
    # Fitting
    coefs = poly.polyfit(xdata, ydata, PolyRank)
    ffit_HR = poly.polyval(x_HR, coefs)
    ffit_Coarse = poly.polyval(xdata, coefs)
    # Residual
    Dif=np.array(ID_SP)-np.array(ffit_Coarse)
    ratio=(np.array(ID_SP)-np.array(ffit_Coarse))/(np.array(ID_SP)/100)
    # Plot differences
    #print('plotMin = ',str(plotMin))
    if 'dif' in kwargs['plotType']:
        fig = plt.figure(figsize=(12,6))
        plt.plot(Mono_max,ID_SP,marker='x',markersize=5,color='g',linewidth=0,label='data')
        plt.plot(x_HR,ffit_HR,color='b',label='SP-fit')
        plt.plot(xdata,Dif*100,marker='x',color='r',label='Dif x100')
        plt.plot(xdata,ratio*1000,marker='x',color='g',label='Difx1000/ID')
        plt.ylabel('ID SP')
        plt.xlabel('Mono')
        plt.xlim(plotMin,plotMax)
        plt.legend(ncol=2, shadow=True, title=mytitle, fancybox=True)
        plt.grid(linestyle='-', linewidth='0.5', color='grey')
        plt.show()
    # Plot raw data + fit
    if 'fit' in kwargs['plotType']:
        fig = plt.figure(figsize=(12,6))
        a1 = fig.add_axes([0,0,1,1])
        a1.plot(xdata+Dif,zdata,marker='*',markersize=10,color='r',linewidth=0,label='Interpolated ID SP')
        for i in range(first_scan,last_scan+1):
            x,y,x_name,y_name=mda_1D(i,det,1,0,filepath,prefix)
            a1.plot(x,y,color='b')
        a1.set_xlim(plotMin,plotMax)
        a1.set(xlabel='Mono')
        a1.set(ylabel='ID SP')
        a1.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
        a1.grid(linestyle='-', linewidth='0.5', color='grey')
        plt.legend(ncol=2, shadow=True, title=mytitle, fancybox=True)
        plt.show()
    # Plot flux curves:
    fdata_x=xdata+Dif
    if 'flux' in kwargs['plotType']:
        fig = plt.figure(figsize=(12,6))
        a1 = fig.add_axes([0,0,1,1])
        a1.plot(fdata_x,fdata,color='r',linewidth=1,label='Flux curves')
        a1.set_xlim(plotMin,plotMax)
        a1.set(xlabel='Mono')
        a1.set(ylabel='ID SP')
        a1.set_yscale('log')
        #a1.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
        a1.grid(linestyle='-', linewidth='0.5', color='grey')
        plt.legend(ncol=2, shadow=True, title=mytitle, fancybox=True)
        plt.show()
    return coefs, fdata_x,fdata 






def read_id_files(first,last,filepath=None,prefix=None,q=True):
    """
    Reads extra PVs
    Return a list of [[ScanNum,ID_SP,grt,mode],...] for mda files between first and last.
    """
    if filepath == None:
        filepath=iex.BL.mda.filepath()
    if prefix == None:
        prefix=iex.BL.mda.prefix()[:-1]
    elif prefix[-1] == "_":
        prefix=prefix[:-1]
    mydata=mdaFile(first,last,filepath=filepath,prefix=prefix,q=q)
    value=[]
    modeDict={'CW,  RCP':'RCP','CCW, LCP':'LCP','H':'H','V':'V'}
    grtDict={1200:'MEG',2400:'HEG'}
    for mdaNum in mydata.scanList:
        if first<=mdaNum<=last:
            extraPVs=mydata.header[mdaNum].all
            try:
                ID=round(extraPVs['ID29:EnergySet.VAL'][2][0]*1000,2)
                mode=modeDict[extraPVs['ID29:ActualMode'][2]]                # extraPVs return 'CW, RCP'
                grt=grtDict[extraPVs['29idmono:GRT_DENSITY'][2][0]]        # extraPVs return 1200
            except KeyError:
                ID=round(extraPVs[b'ID29:EnergySet.VAL'][2][0]*1000,2)
                mode=modeDict[str(extraPVs[b'ID29:ActualMode'][2])[2:-1]]
                grt=grtDict[extraPVs[b'29idmono:GRT_DENSITY'][2][0]]        # extraPVs return 1200
            if len(value)>0 and value[-1][1:] == [ID,grt,mode]:
                pass
            else:
                value.append([mdaNum,ID,grt,mode])
    return value



def id2num(ID,grt,mode,first=0,last=inf,ignore=[],filepath=None,prefix=None,q=True):   # Not efficient - requires to read all 600 files everytime
    """
    Return ScanNum corresponding to a given ID_SP from ExtraPVs (using mdaFile)
    """
    if filepath == None:
        filepath=BL.mda.filepath()
    if prefix == None:
        prefix=BL.mda.prefix()[:-1]
    elif prefix[-1] == "_":
        prefix=prefix[:-1]
    ScanNum = 0
    data_list = read_id_files(first,last,filepath,prefix,q)
    data_short=[x for x in data_list if x[2:]==[grt,mode] and x[0] not in ignore]
    step=data_short[1][1]-data_short[0][1]
    precision=int(step/2)
    ID1=ID-precision
    ID2=ID+precision
    ScanNum=[x[0] for x in data_short if ID1<= x[1]<= ID2]
    if len(ScanNum)==0: result=None
    else: result=ScanNum[0]
    return result



def num2id(ScanNum,grt,mode,first=0,last=inf,ignore=[],filepath=None,prefix=None,q=True):  # Not efficient - requires to read all 600 files everytime
    """
    Return ID SP corresponding to a given ScanNum from ExtraPVs (using mdaFile)
    """
    if filepath == None:
        filepath=BL.mda.filepath()
    if prefix == None:
        prefix=BL.mda.prefix()[:-1]
    elif prefix[-1] == "_":
        prefix=prefix[:-1]
    ID = 0
    data_short=[]
    data_list = read_id_files(first,last,filepath,prefix,q)
    data_short=[x for x in data_list if x[2:]==[grt,mode] and x[0] not in ignore]
    ID=[x[1] for x in data_short if x[0] == ScanNum]
    return ID[0]



def extract_id(first,last,ignore=[],breakpts={'RCP':[600],'H':[400,600],'V':[600],'LCP':[600],'HEG':[],'MEG':[2200,2475]},filepath=None,prefix=None,q=True):
    """
    Breaksdown the info from a calibration files into grt & mode with the corresponding breakpoints (hardcoded):
        [[first, last, 'HEG', 'RCP', [600]],
         [first, last, 'HEG', 'H', [400, 600]]...]  
    """
    if filepath == None:
        filepath=BL.mda.filepath()
    if prefix == None:
        prefix=BL.mda.prefix()[:-1]
    elif prefix[-1] == "_":
        prefix=prefix[:-1]
    IDlog=read_id_files(first,last,filepath,prefix,q)
    #breakpts={'RCP':[600],'H':[400,600],'V':[600],'LCP':[600],'HEG':[],'MEG':[2200]}
    #breakpts={'RCP':[600],'H':[400,600],'V':[600],'LCP':[600],'HEG':[],'MEG':[2200]}
    #breakpts={'RCP':[600],'H':[600],'V':[600],'LCP':[600],'HEG':[],'MEG':[2200,2475]}  # FR changed H breakpoints, missing low energy scans 06/14/2021
    data=[]
    for grt in ['HEG','MEG']:
        for mode in ['RCP','H','V','LCP']:
            tmp=[x for x in IDlog if x[2:]==[grt,mode] and x[0] not in ignore]
            #print(tmp)
            if len(tmp)>0:
                tmp=[tmp[0][0],tmp[-1][0],grt,mode,breakpts[mode]+breakpts[grt]]
            data.append(tmp)
    return data



def update_id_dict(first,last,det,update_file,ignore,**kwargs):
    """
    Calculate new calibration curve for a full set of data.'
    But what if the set is not complete? To be addressed by Future Me
    If update_file == True (ONLY):
    \tupdate the ID dictionary '/home/beams22/29IDUSER/Documents/User_Macros/Macros_29id/IEX_Dictionaries/Dict_IDCal.txt'
    """ 

    if update_file == True:
        foo=input('\nAre you sure you want to update the ID calibration dictionary?\n>')
        if foo == 'Y' or foo == 'y' or foo == 'yes'or foo == 'YES' or foo == 'Yes':
            foo = 'yes'
            print('\nWill save new dictionary to: \'/home/beams22/29IDUSER/Documents/User_Macros/Macros_29id/IEX_Dictionaries/Dict_IDCal.txt\'\n')
        else:
            print('\nCalibration curves will not be saved.\n')
    else:
        print('\nCalibration curves will not be saved.\n')

    kwargs.setdefault('countby',1)
    kwargs.setdefault('mytitle','')
    kwargs.setdefault('plotMin',225)
    kwargs.setdefault('plotMax',3000)
    kwargs.setdefault('plotType',['fit','dif','flux'])
    kwargs.setdefault('filepath',iex.BL.mda.filepath())
    kwargs.setdefault('prefix',iex.BL.mda.prefix())
    kwargs.setdefault('q',True)
    kwargs.setdefault('breakpts',{'RCP':[600],'H':[400,600],'V':[600],'LCP':[600],'HEG':[],'MEG':[2200,2475]})


    breakpts=kwargs['breakpts']
    plotType=kwargs['plotType']
    filepath=kwargs['filepath']
    prefix=kwargs['prefix']
    q=kwargs['q']

    ID_data=extract_id(first,last,ignore,breakpts,filepath,prefix,q)

    ##### Extract parameters:

    mode_dict ={'RCP':0,'LCP':1,'V':2,'H':3}
    id_coef={}
    new_id_function={}
    id_flux={}
    flux_dict={}
    id_energy={}
    energy_dict={}
    for mylist in ID_data:
        if len(mylist) > 0:         # in case of incomplete data set
            tmp0,tmp1,tmp2=[],[],[]
            first_scan=mylist[0]
            last_scan=mylist[1]
            scan_list=[[first_scan,last_scan]]
            countby=1
            #det=15
            #det=33
            poly_order=4
            grt=mylist[2]
            mode=mylist[3]
            breakpts_energy=mylist[4]
            breakpts_scan=[]
            mytitle=mode+' - '+grt 
            
            ## generating list of scans depending on breakpoints:
            if breakpts_energy != []:
                scan_list=[]
                for x in breakpts_energy:
                    breakpts_scan.append(id2num(x,grt,mode,first,last,ignore,filepath,prefix,q))
                breakpts_scan = [i for i in breakpts_scan if i]
                print(breakpts_scan)
                for c,x in  enumerate(breakpts_scan):              # can get the number by extracting step size then /2
                    if c == 0:
                        if x == first_scan: pass
                        else: scan_list.append([first_scan,x])
                    if 0 < c < len(breakpts_scan)-1:
                        scan_list.append([breakpts_scan[c-1],x])
                        scan_list.append([x,breakpts_scan[c+1]])
                    if c == len(breakpts_scan)-1 and (c-1) == 0:
                        scan_list.append([breakpts_scan[c-1],x])
                    if c == len(breakpts_scan)-1:
                        scan_list.append([x,last_scan])
            final_list = [i for n, i in enumerate(scan_list) if i not in scan_list[:n]]     # remove doubles
            energy_list=[]
            for x in final_list:
                ID1=num2id(x[0],grt,mode,first,last,ignore,filepath,prefix,q)
                ID2=num2id(x[1],grt,mode,first,last,ignore,filepath,prefix,q)
                energy_list.append([ID1,ID2])
            energy_list_2=energy_list           # we need the final value for plot purposes (max value)
            if grt == 'HEG':
                energy_list_2[-1][-1]=2500          # but we don't want it to be a cutoff for ID_Calc in dict 
            for (FirstLast,ID,cutoff) in zip(final_list,energy_list,energy_list_2):   
                    plt_min=ID[0]-200
                    plt_max=ID[1]+100
                    if grt=='MEG' and ID[1]>2900:
                        poly_order=5
                        plt_min=ID[0]-200
                    print('det =',det)
                    print('poly_order =',poly_order)
                    
                    subkwargs=kwargs # kwargs contains plotType, filepath, prefix and q
                    newkwargs={'mytitle':mytitle,'countby':countby,"plotMin":plt_min,'plotMax':plt_max}
                    subkwargs.update(newkwargs)
                    #print(newkwargs)
                    #print(FirstLast[0],FirstLast[1])
                    #result,energy_data,flux_data=id_calibration_fit(FirstLast[0],FirstLast[1],det,poly_order,subkwargs)
                    result,energy_data,flux_data=id_calibration_fit(FirstLast[0],FirstLast[1],det,poly_order,mytitle=mytitle,countby=countby,plotMin=plt_min,plotMax=plt_max,plotType=['fit','dif'],filepath=filepath,prefix=prefix)
                    tmp0.append([cutoff[1],result.tolist()])
                    tmp1.append([cutoff[1],energy_data])
                    tmp2.append([cutoff[1],flux_data])
            
            if 2900<tmp0[-1][0]<3000: tmp0[-1][0]=3000
            id_coef[mode_dict[mode]]=tmp0      # dictionary that
            new_id_function[grt]=id_coef      # dictionary containing all the calibration curves forthe data set
            id_energy[mode_dict[mode]]=tmp1      # dictionary that
            energy_dict[grt]=id_energy    
            id_flux[mode_dict[mode]]=tmp2      # dictionary that
            flux_dict[grt]=id_flux

    ##### Read & update old dictionary:

    try:
        id_function=read_dict('Dict_IDCal.txt')
        print(id_function)
        id_function.update(new_id_function)

    except KeyError:
        print("Unable to read previous dictionary")


    if update_file == True and foo == 'yes':
        filepath = "/home/beams22/29IDUSER/Documents/User_Macros/Macros_29id/IEX_Dictionaries/"
        filename = 'Dict_IDCal.txt'

        with open(join(filepath, filename), "a+") as f:   
            f.write('\n======= '+today()+': \n')
            f.write(str(id_function))  
            f.write('\n')
            print('\nWriting dictionary to:',join(filepath, filename))

    if update_file == True and foo == 'yes':
        filepath = "/home/beams22/29IDUSER/Documents/User_Macros/Macros_29id/IEX_Dictionaries/"
        filename = 'Flux_Curves.txt'

        with open(join(filepath, filename), "a+") as f:   
            f.write('\n======= '+today()+': \n')
            f.write('\n----- flux_x:\n')
            f.write(str(energy_dict))  
            f.write('\n----- flux_y:\n')
            f.write(str(flux_dict))  
            f.write('\n')
            print('\nWriting flux curves to:',join(filepath, filename))

    return id_function,energy_dict,flux_dict
